__title__ = 'compas_fab'
__description__ = 'Robotic fabrication package for the COMPAS Framework'
__url__ = 'https://github.com/compas-dev/compas_fab'
__version__ = '0.18.2'
__author__ = 'Gramazio Kohler Research'
__author_email__ = 'gramaziokohler@arch.ethz.ch'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2018 Gramazio Kohler Research'
